package p08_funkcje;

import java.util.Optional;
import java.util.OptionalInt;

public class WartosciOpcjonalne {
    // Gdy funkcja czasami ma określony wynik, ale czasami nie
    // (typowe dla wyszukiwania...)
    // to istnieje kilka sposobów zwracania takiego wyniku.
    
    // Przykładowe funkcje:
    // - zwracanie pierwszego znalezionego elementu parzystego (może żadnego nie być)
    // - max - co zrobić, gdy tablica jest zupełnie pusta?
    
    // Podejście 1 - zwracamy umówioną wartość specjalną
    // Ewentualny problem - jak odróżnić wartość specjalną od "normalnego wyniku"
    // Dla pierwszej parzystej jest to oczywiste, ale dla max już nie - tablica może naprawdę zawierać MIN_INT
    static int pierwszaParzysta_v1(int[] t) {
        for(int e : t) {
            if(e % 2 == 0) {
                return e;
            }
        }
        return -1; // to znaczy że nie znaleźliśmy
    }
    
    static int max_v1(int[] t) {
        int wynik = Integer.MIN_VALUE;
        for(int el : t) {
            if(el > wynik) {
                wynik = el;
            }
        }
        return wynik;
    }
    
    // Podejście drugie: null jako informacja o tym, że wyniku brak
    // nulla można użyć tylko, gdy typ jest typem obiektowym, a nie typem prostym
    // W przypadku liczb musimy zatem przejść na typy Integer, Double itp.
    // W dodatku "z nullami są problemy..."
    
    static Integer pierwszaParzysta_v2(int[] t) {
        for(int e : t) {
            if(e % 2 == 0) {
                return e;
            }
        }
        return null;
    }
    
    static Integer max_v2(int[] t) {
        if(t == null || t.length == 0) {
            return null;
        }
        int wynik = t[0];
        for(int el : t) {
            if(el > wynik) {
                wynik = el;
            }
        }
        return wynik;
    }
    
    // Od Java 8 pojawiły się typy "Optional".
    // Optional<T> to jest typ generyczny dobry dla wszystkich typów obiektowych.
    // Dla liczb dostępne są OptionalInt, OptionalLong, OptionDouble.
    // Funkcja, która "czasami zwraca wynik, a czasami nie" powinna być typu opcjonalnego
    // Od Javy 8 jest to zalecane.
    static OptionalInt pierwszaParzysta_v3(int[] t) {
        for(int e : t) {
            if(e % 2 == 0) {
                return OptionalInt.of(e);
            }
        }
        return OptionalInt.empty();
    }
    
    static OptionalInt max_v3(int[] t) {
        if(t == null || t.length == 0) {
            return OptionalInt.empty();
        }
        int wynik = t[0];
        for(int el : t) {
            if(el > wynik) {
                wynik = el;
            }
        }
        return OptionalInt.of(wynik);
    }
    
    static Optional<String> najdluzszyNapis(String[] t) {
        if(t == null || t.length == 0) {
            return Optional.empty();
        }
        String wynik = "";
        for(String s : t) {
            if(s != null && s.length() > wynik.length()) {
                wynik = s;
            }
        }
        return Optional.of(wynik);
    }
    
    // Informacyjnie: można jeszcze wyrzucać wyjątek, jeśli funkcja nie jest w stanie dać wyniku

    public static void main(String[] args) {
        int[] a = {5, 10, 15, 20};
        int[] b = {5, 3, 15};
        int[] e = {};
        String[] imiona = {"Ala", "Natalia", "Edward", "Zenon", "Agnieszka", "Jan"};
        String[] brakNapisow = {};

        System.out.println("pp_v1:");
        System.out.println(pierwszaParzysta_v1(a));
        System.out.println(pierwszaParzysta_v1(b));
        System.out.println(pierwszaParzysta_v1(e));
        
        System.out.println("\nmax_v1:");
        System.out.println(max_v1(a));
        System.out.println(max_v1(b));
        System.out.println(max_v1(e));

        System.out.println("\npp_v2:");
        System.out.println(pierwszaParzysta_v2(a));
        System.out.println(pierwszaParzysta_v2(b));
        System.out.println(pierwszaParzysta_v2(e));
        
        System.out.println("\nmax_v2:");
        System.out.println(max_v2(a));
        System.out.println(max_v2(b));
        System.out.println(max_v2(e));

        System.out.println("\npp_v3:");
        System.out.println(pierwszaParzysta_v3(a));
        System.out.println(pierwszaParzysta_v3(b));
        System.out.println(pierwszaParzysta_v3(e));
        
        System.out.println("\nmax_v3:");
        System.out.println(max_v3(a));
        System.out.println(max_v3(b));
        System.out.println(max_v3(e));
        System.out.println();
        
        System.out.println(najdluzszyNapis(imiona));
        System.out.println(najdluzszyNapis(brakNapisow));
        
        // sprawdzanie, czy Optional jest niepusty
        Optional<String> imie = najdluzszyNapis(imiona);
        if(imie.isPresent()) {
            System.out.println("Najdłuższe imię to " + imie.get());
        } else {
            System.out.println("Brak imion");
            
        }
        Optional<String> napis = najdluzszyNapis(brakNapisow);
        if(napis.isPresent()) {
            System.out.println("Najdłuższy napis to " + napis.get());
        } else {
            System.out.println("Brak napisów");
        }
        System.out.println();
        
        // można też stosować zapisy "funkcyjne"
        najdluzszyNapis(imiona).ifPresent(im -> System.out.println("Znalazłem imię " + im));
        
        najdluzszyNapis(imiona)
                .map(String::toUpperCase)
                .ifPresent(System.out::println);
        
        // wersja z else:
        pierwszaParzysta_v3(b).ifPresentOrElse(
                y -> System.out.println("Pierwsza parzysta to " + y),
                () -> System.out.println("Nie znaleziono parzystej"));
        
    }

}
